import { GenericDataError } from "./../helpers/errors";
import Joi from "@hapi/joi";
import { Marshaller } from "@aws/dynamodb-auto-marshaller";
import { getJoiValidationErrors } from "../core/base-joi-helper";
import BaseMixins from "../core/base-mixins";
import { coreSchemaDefinition, } from "./base-schema";
import { DynamoManageTable } from "./dynamo-manage-table";
function createTenantSchema(schemaMapDef) {
    return Joi.object().keys({ ...schemaMapDef, ...coreSchemaDefinition });
}
export default class DynamoDataOperation extends BaseMixins {
    constructor({ dynamoDb, schemaDef, dynamoDbClient, secondaryIndexOptions, featureIdentityValue, baseTableName, strictRequiredFields, dataKeyGenerator, }) {
        super();
        Object.defineProperty(this, "here_partitionKeyFieldName", {
            enumerable: true,
            configurable: true,
            writable: true,
            value: "id"
        });
        Object.defineProperty(this, "here_sortKeyFieldName", {
            enumerable: true,
            configurable: true,
            writable: true,
            value: "featureIdentity"
        });
        Object.defineProperty(this, "here_dynamoDbClient", {
            enumerable: true,
            configurable: true,
            writable: true,
            value: void 0
        });
        Object.defineProperty(this, "here_dynamoDb", {
            enumerable: true,
            configurable: true,
            writable: true,
            value: void 0
        });
        Object.defineProperty(this, "here_dataKeyGenerator", {
            enumerable: true,
            configurable: true,
            writable: true,
            value: void 0
        });
        Object.defineProperty(this, "here_schema", {
            enumerable: true,
            configurable: true,
            writable: true,
            value: void 0
        });
        Object.defineProperty(this, "here_marshaller", {
            enumerable: true,
            configurable: true,
            writable: true,
            value: void 0
        });
        Object.defineProperty(this, "here_tableFullName", {
            enumerable: true,
            configurable: true,
            writable: true,
            value: void 0
        });
        Object.defineProperty(this, "here_strictRequiredFields", {
            enumerable: true,
            configurable: true,
            writable: true,
            value: void 0
        });
        Object.defineProperty(this, "here_featureIdentityValue", {
            enumerable: true,
            configurable: true,
            writable: true,
            value: void 0
        });
        Object.defineProperty(this, "here_secondaryIndexOptions", {
            enumerable: true,
            configurable: true,
            writable: true,
            value: void 0
        });
        Object.defineProperty(this, "here_tableManager", {
            enumerable: true,
            configurable: true,
            writable: true,
            value: void 0
        });
        this.here_dynamoDb = dynamoDb;
        this.here_dynamoDbClient = dynamoDbClient;
        this.here_dataKeyGenerator = dataKeyGenerator;
        this.here_schema = createTenantSchema(schemaDef);
        this.here_tableFullName = baseTableName;
        this.here_marshaller = new Marshaller({ onEmpty: "omit" });
        this.here_featureIdentityValue = featureIdentityValue;
        this.here_secondaryIndexOptions = secondaryIndexOptions;
        this.here_strictRequiredFields = strictRequiredFields;
    }
    getTableManager() {
        if (!this.here_tableManager) {
            this.here_tableManager = new DynamoManageTable({
                dynamoDb: this.here_dynamoDb,
                secondaryIndexOptions: this.here_secondaryIndexOptions,
                tableFullName: this.here_tableFullName,
                partitionKeyFieldName: this.here_partitionKeyFieldName,
                sortKeyFieldName: this.here_sortKeyFieldName,
            });
        }
        return this.here_tableManager;
    }
    _dynamoDbClient() {
        return this.here_dynamoDbClient();
    }
    _dynamoDb() {
        return this.here_dynamoDb();
    }
    generateDynamoTableKey() {
        return this.here_dataKeyGenerator();
    }
    _getLocalVariables() {
        return {
            partitionKeyFieldName: this.here_partitionKeyFieldName,
            sortKeyFieldName: this.here_sortKeyFieldName,
            featureIdentityValue: this.here_featureIdentityValue,
            tableFullName: this.here_tableFullName,
            secondaryIndexOptions: this.here_secondaryIndexOptions,
            strictRequiredFields: this.here_strictRequiredFields,
        };
    }
    _getBaseObject({ dataId }) {
        const { partitionKeyFieldName, sortKeyFieldName, featureIdentityValue, } = this._getLocalVariables();
        const dataMust = {
            [partitionKeyFieldName]: dataId,
            [sortKeyFieldName]: featureIdentityValue,
        };
        return dataMust;
    }
    _checkValidateMustBeAnObjectDataType(data) {
        if (!data || typeof data !== "object") {
            throw new GenericDataError(`Data MUST be valid object`);
        }
    }
    _checkValidateStrictRequiredFields(onDataObj) {
        this._checkValidateMustBeAnObjectDataType(onDataObj);
        const { strictRequiredFields } = this._getLocalVariables();
        if (strictRequiredFields?.length) {
            for (const field of strictRequiredFields) {
                if (onDataObj[field] === null || onDataObj[field] === undefined) {
                    throw new GenericDataError(`Required field NOT defined`);
                }
            }
        }
    }
    async allCreateOneBase({ data }) {
        this._checkValidateStrictRequiredFields(data);
        const { tableFullName, partitionKeyFieldName } = this._getLocalVariables();
        let dataId = data[partitionKeyFieldName];
        if (!dataId) {
            dataId = this.generateDynamoTableKey();
        }
        const baseData = {
            createdAtDate: new Date().toISOString(),
        };
        const dataMust = this._getBaseObject({ dataId });
        const fullData = { ...data, ...baseData, ...dataMust };
        const { validatedData, marshalled, } = await this._allHelpValidateMarshallAndGetValue(fullData);
        const params = {
            TableName: tableFullName,
            Item: marshalled,
        };
        await this._dynamoDb().putItem(params).promise();
        const result = validatedData;
        return result;
    }
    withConditionPassed({ item, withCondition, }) {
        if (item && withCondition?.length) {
            const isPassed = withCondition.every(({ field, equals }) => {
                return item[field] !== undefined && item[field] === equals;
            });
            return isPassed;
        }
        return true;
    }
    async allGetOneByIdBase({ dataId, withCondition, }) {
        const { partitionKeyFieldName, sortKeyFieldName, featureIdentityValue, tableFullName, } = this._getLocalVariables();
        this.allHelpValidateRequiredString({
            QueryGetOnePartitionKey: dataId,
            QueryGetOneSortKey: featureIdentityValue,
        });
        const params = {
            TableName: tableFullName,
            Key: {
                [partitionKeyFieldName]: dataId,
                [sortKeyFieldName]: featureIdentityValue,
            },
        };
        const result = await this._dynamoDbClient().get(params).promise();
        const item = result.Item;
        if (!item) {
            return null;
        }
        const isPassed = this.withConditionPassed({ withCondition, item });
        if (!isPassed) {
            return null;
        }
        return item;
    }
    async allUpdateOneDirectBase({ data }) {
        this._checkValidateStrictRequiredFields(data);
        const { tableFullName, partitionKeyFieldName } = this._getLocalVariables();
        const dataId = data[partitionKeyFieldName];
        if (!dataId) {
            throw new GenericDataError("Update data requires sort key field value");
        }
        const dataMust = this._getBaseObject({ dataId });
        const baseData = {
            lastModifiedDate: new Date().toISOString(),
        };
        const fullData = { ...data, ...baseData, ...dataMust };
        const { validatedData, marshalled, } = await this._allHelpValidateMarshallAndGetValue(fullData);
        const params = {
            TableName: tableFullName,
            Item: marshalled,
        };
        await this._dynamoDb().putItem(params).promise();
        const result = validatedData;
        return result;
    }
    async allUpdateOneByIdBase({ dataId, data, withCondition, }) {
        this._checkValidateStrictRequiredFields(data);
        const { tableFullName, partitionKeyFieldName } = this._getLocalVariables();
        this.allHelpValidateRequiredString({ Update1DataId: dataId });
        const dataInDb = await this.allGetOneByIdBase({ dataId });
        if (!(dataInDb && dataInDb[partitionKeyFieldName])) {
            throw new GenericDataError("Data does NOT exists");
        }
        const isPassed = this.withConditionPassed({
            withCondition,
            item: dataInDb,
        });
        if (!isPassed) {
            throw new GenericDataError("Update condition failed");
        }
        const dataMust = this._getBaseObject({
            dataId: dataInDb[partitionKeyFieldName],
        });
        const fullData = { ...dataInDb, ...data, ...dataMust };
        const { validatedData, marshalled, } = await this._allHelpValidateMarshallAndGetValue(fullData);
        const params = {
            TableName: tableFullName,
            Item: marshalled,
        };
        await this._dynamoDb().putItem(params).promise();
        const result = validatedData;
        return result;
    }
    async _allHelpValidateMarshallAndGetValue(data) {
        const { error, value } = this.here_schema.validate(data, {
            stripUnknown: true,
        });
        if (error) {
            const msg = getJoiValidationErrors(error) ?? "Validation error occured";
            throw new GenericDataError(msg);
        }
        const marshalledData = this.here_marshaller.marshallItem(value);
        return await Promise.resolve({
            validatedData: value,
            marshalled: marshalledData,
        });
    }
    async allQueryGetManyByConditionBase(paramOptions) {
        paramOptions.pagingParams = undefined;
        const result = await this.allQueryGetManyByConditionPaginateBase(paramOptions);
        if (result?.mainResult?.length) {
            return result.mainResult;
        }
        return [];
    }
    async allQueryGetManyByConditionPaginateBase(paramOptions) {
        const { tableFullName, sortKeyFieldName, partitionKeyFieldName, } = this._getLocalVariables();
        if (!paramOptions?.partitionKeyQuery?.equals === undefined) {
            throw new GenericDataError("Invalid Hash key value");
        }
        if (!sortKeyFieldName) {
            throw new GenericDataError("Bad query sort configuration");
        }
        let sortKeyQuery = {};
        const sortKeyQueryData = paramOptions.sortKeyQuery;
        if (sortKeyQueryData) {
            if (sortKeyQueryData[sortKeyFieldName]) {
                sortKeyQuery = {
                    [sortKeyFieldName]: sortKeyQueryData[sortKeyFieldName],
                };
            }
            else {
                throw new GenericDataError("Invalid Sort key value");
            }
        }
        const filterHashSortKey = this.__helperDynamoFilterOperation({
            queryDefs: {
                ...sortKeyQuery,
                ...{
                    [partitionKeyFieldName]: paramOptions.partitionKeyQuery.equals,
                },
            },
            projectionFields: paramOptions?.fields ?? undefined,
        });
        let otherFilterExpression = undefined;
        let otherExpressionAttributeValues = undefined;
        let otherExpressionAttributeNames = undefined;
        if (paramOptions?.query) {
            const filterOtherAttr = this.__helperDynamoFilterOperation({
                queryDefs: paramOptions.query,
                projectionFields: null,
            });
            otherExpressionAttributeValues =
                filterOtherAttr.expressionAttributeValues;
            otherExpressionAttributeNames = filterOtherAttr.expressionAttributeNames;
            if (filterOtherAttr?.filterExpression &&
                filterOtherAttr?.filterExpression.length > 1) {
                otherFilterExpression = filterOtherAttr.filterExpression;
            }
        }
        const params = {
            TableName: tableFullName,
            KeyConditionExpression: filterHashSortKey.filterExpression,
            ExpressionAttributeValues: {
                ...otherExpressionAttributeValues,
                ...filterHashSortKey.expressionAttributeValues,
            },
            FilterExpression: otherFilterExpression ?? undefined,
            ExpressionAttributeNames: {
                ...otherExpressionAttributeNames,
                ...filterHashSortKey.expressionAttributeNames,
            },
        };
        if (filterHashSortKey?.projectionExpressionAttr) {
            params.ProjectionExpression = filterHashSortKey.projectionExpressionAttr;
        }
        if (paramOptions?.pagingParams?.orderDesc === true) {
            params.ScanIndexForward = false;
        }
        const hashKeyAndSortKey = [
            partitionKeyFieldName,
            sortKeyFieldName,
        ];
        const paginationObjects = { ...paramOptions.pagingParams };
        const result = await this.__helperDynamoQueryProcessor({
            dynamoDbClient: () => this._dynamoDbClient(),
            params,
            hashKeyAndSortKey,
            ...paginationObjects,
        });
        return result;
    }
    async allBatchGetManyByIdsBase({ dataIds, fields, withCondition, }) {
        dataIds.forEach((dataId) => {
            this.allHelpValidateRequiredString({
                BatchGetDataId: dataId,
            });
        });
        const getRandom = () => [
            "rand",
            Math.round(Math.random() * 99999),
            Math.round(Math.random() * 88888),
            Math.round(Math.random() * 99),
        ].join("");
        return new Promise((resolve, reject) => {
            const { tableFullName, partitionKeyFieldName, sortKeyFieldName, featureIdentityValue, } = this._getLocalVariables();
            const getArray = dataIds.map((dataId) => {
                const params01 = {
                    [partitionKeyFieldName]: { S: dataId },
                    [sortKeyFieldName]: { S: featureIdentityValue },
                };
                return params01;
            });
            let projectionExpression = undefined;
            let expressionAttributeNames = undefined;
            if (fields?.length) {
                const _fields = [...fields];
                if (fields?.length && withCondition?.length) {
                    withCondition.forEach((condition) => {
                        if (!fields.includes(condition.field)) {
                            _fields.push(condition.field);
                        }
                    });
                }
                expressionAttributeNames = {};
                _fields.forEach((fieldName) => {
                    if (typeof fieldName === "string") {
                        if (expressionAttributeNames) {
                            const attrKeyHash = `#attrKey${getRandom()}k`.toLowerCase();
                            expressionAttributeNames[attrKeyHash] = fieldName;
                        }
                    }
                });
                if (Object.keys(expressionAttributeNames)?.length) {
                    projectionExpression = Object.keys(expressionAttributeNames).join(",");
                }
                else {
                    projectionExpression = undefined;
                    expressionAttributeNames = undefined;
                }
            }
            const params = {
                RequestItems: {
                    [tableFullName]: {
                        Keys: [...getArray],
                        ConsistentRead: true,
                        ProjectionExpression: projectionExpression,
                        ExpressionAttributeNames: expressionAttributeNames,
                    },
                },
            };
            let returnedItems = [];
            const resolveItemResults = (resultItems) => {
                if (resultItems?.length && withCondition?.length) {
                    return resultItems.filter((item) => {
                        return withCondition.every((condition) => {
                            return item[condition.field] === condition.equals;
                        });
                    });
                }
                return resultItems;
            };
            const batchGetUntilDone = (err, data) => {
                if (err) {
                    if (returnedItems?.length) {
                        resolve(resolveItemResults(returnedItems));
                    }
                    else {
                        reject(err.stack);
                    }
                }
                else {
                    if (data?.Responses) {
                        const itemListRaw = data.Responses[tableFullName];
                        const itemList = itemListRaw.map((item) => {
                            return this.here_marshaller.unmarshallItem(item);
                        });
                        returnedItems = [...returnedItems, ...itemList];
                    }
                    if (data?.UnprocessedKeys) {
                        const _params = {
                            RequestItems: data.UnprocessedKeys,
                        };
                        console.log({ dynamoBatchGetParams: _params });
                        this._dynamoDb().batchGetItem(_params, batchGetUntilDone);
                    }
                    else {
                        resolve(resolveItemResults(returnedItems));
                    }
                }
            };
            this._dynamoDb().batchGetItem(params, batchGetUntilDone);
        });
    }
    async allQuerySecondaryIndexBase(paramOption) {
        paramOption.pagingParams = undefined;
        const result = await this.allQuerySecondaryIndexPaginateBase(paramOption);
        if (result?.mainResult) {
            return result.mainResult;
        }
        return [];
    }
    async allQuerySecondaryIndexPaginateBase(paramOption) {
        const { tableFullName, secondaryIndexOptions } = this._getLocalVariables();
        if (!secondaryIndexOptions?.length) {
            throw new GenericDataError("Invalid secondary index definitions");
        }
        const { indexName, partitionKeyQuery, sortKeyQuery, fields, pagingParams, query, } = paramOption;
        const secondaryIndex = secondaryIndexOptions.find((item) => {
            return item.indexName === indexName;
        });
        if (!secondaryIndex) {
            throw new GenericDataError("Secondary index not named/defined");
        }
        const partitionKeyFieldName = secondaryIndex.keyFieldName;
        const sortKeyFieldName = secondaryIndex.sortFieldName;
        const partitionSortKeyQuery = sortKeyQuery
            ? {
                ...{ [sortKeyFieldName]: sortKeyQuery },
                ...{ [partitionKeyFieldName]: partitionKeyQuery.equals },
            }
            : { [partitionKeyFieldName]: partitionKeyQuery.equals };
        const { expressionAttributeValues, filterExpression, projectionExpressionAttr, expressionAttributeNames, } = this.__helperDynamoFilterOperation({
            queryDefs: partitionSortKeyQuery,
            projectionFields: fields ?? undefined,
        });
        let otherFilterExpression = undefined;
        let otherExpressionAttributeValues = undefined;
        let otherExpressionAttributeNames = undefined;
        if (query) {
            const otherAttr = this.__helperDynamoFilterOperation({
                queryDefs: query,
                projectionFields: null,
            });
            otherExpressionAttributeValues = otherAttr.expressionAttributeValues;
            otherExpressionAttributeNames = otherAttr.expressionAttributeNames;
            if (otherAttr?.filterExpression?.length &&
                otherAttr?.filterExpression.length > 1) {
                otherFilterExpression = otherAttr.filterExpression;
            }
        }
        const params = {
            TableName: tableFullName,
            IndexName: indexName,
            KeyConditionExpression: filterExpression,
            ExpressionAttributeValues: {
                ...otherExpressionAttributeValues,
                ...expressionAttributeValues,
            },
            FilterExpression: otherFilterExpression ?? undefined,
            ExpressionAttributeNames: {
                ...otherExpressionAttributeNames,
                ...expressionAttributeNames,
            },
        };
        const orderDesc = pagingParams?.orderDesc === true;
        if (orderDesc) {
            params.ScanIndexForward = false;
        }
        if (projectionExpressionAttr) {
            params.ProjectionExpression = projectionExpressionAttr;
        }
        const hashKeyAndSortKey = [
            partitionKeyFieldName,
            sortKeyFieldName,
        ];
        const result = await this.__helperDynamoQueryProcessor({
            dynamoDbClient: () => this._dynamoDbClient(),
            params,
            orderDesc,
            hashKeyAndSortKey,
            ...pagingParams,
        });
        return result;
    }
    async allDeleteByIdBase({ dataId, withCondition, }) {
        this.allHelpValidateRequiredString({ Del1SortKey: dataId });
        const { tableFullName, partitionKeyFieldName, sortKeyFieldName, featureIdentityValue, } = this._getLocalVariables();
        const dataExist = await this.allGetOneByIdBase({ dataId, withCondition });
        if (!(dataExist && dataExist[partitionKeyFieldName])) {
            throw new GenericDataError("Record does NOT exists");
        }
        const params = {
            TableName: tableFullName,
            Key: {
                [partitionKeyFieldName]: dataId,
                [sortKeyFieldName]: featureIdentityValue,
            },
        };
        await this._dynamoDbClient().delete(params).promise();
        return dataExist;
    }
    async allDeleteManyDangerouselyByIds({ dataIds, }) {
        dataIds.forEach((sortKeyValue) => {
            this.allHelpValidateRequiredString({
                DelSortKey: sortKeyValue,
            });
        });
        const { tableFullName, partitionKeyFieldName, sortKeyFieldName, featureIdentityValue, } = this._getLocalVariables();
        const delArray = dataIds.map((dataId) => {
            const params01 = {
                DeleteRequest: {
                    Key: {
                        [partitionKeyFieldName]: { S: dataId },
                        [sortKeyFieldName]: { S: featureIdentityValue },
                    },
                },
            };
            return params01;
        });
        const params = {
            RequestItems: {
                [tableFullName]: delArray,
            },
        };
        await this._dynamoDb().batchWriteItem(params).promise();
        return true;
    }
}
